library disposable_service;

import 'dart:collection';

import 'package:flutter/widgets.dart';

/// Interface used in [registerDisposableService]
///
/// [onDispose] will be called when [DisposableServiceManager] is disposed
/// if the instance is registered through [registerDisposableService]
abstract class DisposableService {
  /// Override this method to be called when the state of [DisposableServiceManager] is disposed.
  void onDispose();
}

class _DisposableServiceListEntry
    extends LinkedListEntry<_DisposableServiceListEntry> {
  final DisposableService service;
  final Object scope;

  _DisposableServiceListEntry(
    this.service, {
    this.scope,
  });
}

final _disposableServiceEntries = LinkedList<_DisposableServiceListEntry>();

List<DisposableService> get registeredDisposableServices {
  return _disposableServiceEntries.map((e) => e.service).toList();
}

/// Registers an [DisposableService] instance to be disposed by [DisposableServiceManager].
void registerDisposableService(DisposableService service, {Object scope}) {
  assert(service != null);
  _disposableServiceEntries.add(
    _DisposableServiceListEntry(service, scope: scope),
  );
}

/// Manages to call onDispose of registered [DisposableService] when this widget's state is disposed.
///
/// If [scope] is specified, this will only call onDispose of the [DisposableService] registered under the same [scope] in [registeredDisposableServices]
class DisposableServiceManager extends StatefulWidget {
  /// child widget. This must not be null
  final Widget child;

  /// specify this when you only want to call onDispose of [DisposableService] registed under the same scope
  final Object scope;

  const DisposableServiceManager({
    Key key,
    @required this.child,
    this.scope,
  })  : assert(child != null),
        super(key: key);

  @override
  _DisposableServiceManagerState createState() =>
      _DisposableServiceManagerState();
}

class _DisposableServiceManagerState extends State<DisposableServiceManager> {
  @override
  void dispose() {
    super.dispose();
    final entriesToDispose = _disposableServiceEntries
        .where((entry) => widget.scope == null || entry.scope == widget.scope)
        .toList();
    for (final entry in entriesToDispose) {
      entry.service?.onDispose();
      entry.unlink();
    }
  }

  @override
  Widget build(BuildContext context) {
    return widget.child;
  }
}
