import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';

import 'package:disposable_service/disposable_service.dart';
import 'package:mockito/mockito.dart';

class DisposableServiceImplMock extends Mock implements DisposableService {}

void main() {
  testWidgets(
    'onDispose is not called if DisposableService is not registered',
    (tester) async {
      final disposableService = DisposableServiceImplMock();

      await tester.pumpWidget(DisposableServiceManager(child: Container()));
      await tester.pumpWidget(Container());

      verifyNever(disposableService.onDispose());
    },
  );

  testWidgets(
    'onDispose is called if DisposableService is registered',
    (tester) async {
      final disposableService = DisposableServiceImplMock();

      registerDisposableService(disposableService);
      await tester.pumpWidget(DisposableServiceManager(child: Container()));
      await tester.pumpWidget(Container());

      verify(disposableService.onDispose()).called(1);
    },
  );

  testWidgets(
    'onDispose is called for all registered DisposableService after DisposableServiceManger with no scope is specified is disposed',
    (tester) async {
      final disposableService1 = DisposableServiceImplMock();
      registerDisposableService(disposableService1);
      final disposableService2 = DisposableServiceImplMock();
      registerDisposableService(disposableService2, scope: 'scope1');
      final disposableService3 = DisposableServiceImplMock();
      registerDisposableService(disposableService3, scope: 'scope2');

      await tester.pumpWidget(DisposableServiceManager(child: Container()));
      await tester.pumpWidget(Container());

      verify(disposableService1.onDispose()).called(1);
      verify(disposableService2.onDispose()).called(1);
      verify(disposableService3.onDispose()).called(1);
    },
  );

  testWidgets(
    'onDispose is called only for the DisposableService registered for specified scope after DisposableServiceManger which has the specified scope is disposed',
    (tester) async {
      final disposableService1 = DisposableServiceImplMock();
      registerDisposableService(disposableService1);
      final disposableService2 = DisposableServiceImplMock();
      registerDisposableService(disposableService2, scope: 'scope1');
      final disposableService3 = DisposableServiceImplMock();
      registerDisposableService(disposableService3, scope: 'scope2');

      await tester.pumpWidget(DisposableServiceManager(
        child: Container(),
        scope: 'scope1',
      ));
      await tester.pumpWidget(Container());

      verifyNever(disposableService1.onDispose());
      verify(disposableService2.onDispose()).called(1);
      verifyNever(disposableService3.onDispose());
    },
  );

  testWidgets(
    'all registered DisposableServices are dereferenced after DisposableServiceManger which has no scope specified is disposed',
    (tester) async {
      for (final service in List<DisposableService>.generate(
        10,
        (_) => DisposableServiceImplMock(),
      )) {
        registerDisposableService(service);
      }

      await tester.pumpWidget(DisposableServiceManager(child: Container()));
      await tester.pumpWidget(Container());

      expect(registeredDisposableServices, isEmpty);
    },
  );

  testWidgets(
    'only the DisposableService registered for specified scope is dereferenced after DisposableServiceManger which has the specified scope is disposed',
    (tester) async {
      final disposableService1 = DisposableServiceImplMock();
      registerDisposableService(disposableService1);
      final disposableService2 = DisposableServiceImplMock();
      registerDisposableService(disposableService2, scope: 'scope1');
      final disposableService3 = DisposableServiceImplMock();
      registerDisposableService(disposableService3, scope: 'scope2');

      await tester.pumpWidget(DisposableServiceManager(
        child: Container(),
        scope: 'scope1',
      ));
      await tester.pumpWidget(Container());

      expect(
        registeredDisposableServices,
        containsAll([disposableService1, disposableService3]),
      );
      expect(
        registeredDisposableServices.contains(disposableService2),
        isFalse,
      );
    },
  );
}
